/*************** <auto-copyright.pl BEGIN do not edit this line> **************
 *
 * osgBullet is (C) Copyright 2009 by Kenneth Mark Bryden
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 *************** <auto-copyright.pl END do not edit this line> ***************/

#include "wiz.h"
#include "utilities.h"
#include "wizmain.h"

#include <osgDB/ReadFile>
#include <osgViewer/Viewer>
#include <osg/MatrixTransform>
#include <osg/ShapeDrawable>
#include <osg/Geode>

#include <osgbBullet/MotionState.h>
#include <osgbBullet/CollisionShapes.h>
#include <osgbBullet/Utils.h>

#include <btBulletDynamicsCommon.h>

#include <string>
#include <osg/io_utils>

btDynamicsWorld* dynamicsWorld;
btDefaultCollisionConfiguration * collisionConfiguration;
btCollisionDispatcher * dispatcher;
btConstraintSolver * solver;
btBroadphaseInterface * inter;

struct ObjectPhysic
{
btCollisionShape* cs;
osgbBullet::MotionState* motion;
btRigidBody* body;
};

#define DICE_COUNT 2
struct ObjectPhysic diceBullet[DICE_COUNT];
osg::MatrixTransform* diceOSG[DICE_COUNT];
struct ObjectPhysic shakeBoxBullet;
osg::MatrixTransform* shakeBoxOSG;

osg::MatrixTransform* makeDie()
{
    osg::MatrixTransform* root = new osg::MatrixTransform;
    osg::Node* node = osgDB::readNodeFile( "dice.osg" );
    root->addChild( node );

    return( root );
}

void initDiePhysic(osg::MatrixTransform* root, struct ObjectPhysic *dp)
{
    dp->cs = osgbBullet::btBoxCollisionShapeFromOSG( root->getChild(0) );
    dp->motion = new osgbBullet::MotionState();
    dp->motion->setTransform( root );
    btScalar mass( 1. );
    btVector3 inertia( 0, 0, 0 );
    dp->cs->calculateLocalInertia( mass, inertia );
    btRigidBody::btRigidBodyConstructionInfo rb( mass, dp->motion, dp->cs, inertia );
    dp->body = new btRigidBody( rb );
    //dp->body->setActivationState( DISABLE_DEACTIVATION );
    dynamicsWorld->addRigidBody( dp->body );
}

osg::Geode* osgBox( const osg::Vec3& center, const osg::Vec3& halfLengths )
{
    osg::Vec3 l( halfLengths * 2. );
    osg::Box* box = new osg::Box( center, l.x(), l.y(), l.z() );
    osg::ShapeDrawable* shape = new osg::ShapeDrawable( box );
    shape->setColor( osg::Vec4( 1., 1., 1., 1. ) );
    osg::Geode* geode = new osg::Geode();
    geode->addDrawable( shape );
    return( geode );
}

float xDim( 10. );
float yDim( 10. );
float zDim( 6. );
float thick( .1 );

osg::MatrixTransform* makeShakeBox()
{
    osg::MatrixTransform* shakeBox = new osg::MatrixTransform;
    { // floor -Z (far back of the shake cube)
        osg::Vec3 halfLengths( xDim*.5, yDim*.5, thick*.5 );
        osg::Vec3 center( 0., 0., zDim*.5 );
        shakeBox->addChild( osgBox( center, halfLengths ) );
    }
    { // top +Z (invisible, to allow user to see through; no OSG analogue
        osg::Vec3 halfLengths( xDim*.5, yDim*.5, thick*.5 );
        osg::Vec3 center( 0., 0., -zDim*.5 );
        //shakeBox->addChild( osgBox( center, halfLengths ) );
    }
    { // left -X
        osg::Vec3 halfLengths( thick*.5, yDim*.5, zDim*.5 );
        osg::Vec3 center( -xDim*.5, 0., 0. );
        shakeBox->addChild( osgBox( center, halfLengths ) );
    }
    { // right +X
        osg::Vec3 halfLengths( thick*.5, yDim*.5, zDim*.5 );
        osg::Vec3 center( xDim*.5, 0., 0. );
        shakeBox->addChild( osgBox( center, halfLengths ) );
    }
    { // bottom of window -Y
        osg::Vec3 halfLengths( xDim*.5, thick*.5, zDim*.5 );
        osg::Vec3 center( 0., -yDim*.5, 0. );
        shakeBox->addChild( osgBox( center, halfLengths ) );
    }
    { // bottom of window -Y
        osg::Vec3 halfLengths( xDim*.5, thick*.5, zDim*.5 );
        osg::Vec3 center( 0., yDim*.5, 0. );
        shakeBox->addChild( osgBox( center, halfLengths ) );
    }
    return shakeBox;
}

void initShakeBoxPhysic()
{
    btCompoundShape *cs = new btCompoundShape;
    shakeBoxBullet.cs = cs;
    { // floor -Z (far back of the shake cube)
        osg::Vec3 halfLengths( xDim*.5, yDim*.5, thick*.5 );
        osg::Vec3 center( 0., 0., zDim*.5 );
        btBoxShape* box = new btBoxShape( osgbBullet::asBtVector3( halfLengths ) );
        btTransform trans; trans.setIdentity();
        trans.setOrigin( osgbBullet::asBtVector3( center ) );
        cs->addChildShape( trans, box );
    }
    { // top +Z (invisible, to allow user to see through; no OSG analogue
        osg::Vec3 halfLengths( xDim*.5, yDim*.5, thick*.5 );
        osg::Vec3 center( 0., 0., -zDim*.5 );
        btBoxShape* box = new btBoxShape( osgbBullet::asBtVector3( halfLengths ) );
        btTransform trans; trans.setIdentity();
        trans.setOrigin( osgbBullet::asBtVector3( center ) );
        cs->addChildShape( trans, box );
    }
    { // left -X
        osg::Vec3 halfLengths( thick*.5, yDim*.5, zDim*.5 );
        osg::Vec3 center( -xDim*.5, 0., 0. );
        btBoxShape* box = new btBoxShape( osgbBullet::asBtVector3( halfLengths ) );
        btTransform trans; trans.setIdentity();
        trans.setOrigin( osgbBullet::asBtVector3( center ) );
        cs->addChildShape( trans, box );
    }
    { // right +X
        osg::Vec3 halfLengths( thick*.5, yDim*.5, zDim*.5 );
        osg::Vec3 center( xDim*.5, 0., 0. );
        btBoxShape* box = new btBoxShape( osgbBullet::asBtVector3( halfLengths ) );
        btTransform trans; trans.setIdentity();
        trans.setOrigin( osgbBullet::asBtVector3( center ) );
        cs->addChildShape( trans, box );
    }
    { // bottom of window -Y
        osg::Vec3 halfLengths( xDim*.5, thick*.5, zDim*.5 );
        osg::Vec3 center( 0., -yDim*.5, 0. );
        btBoxShape* box = new btBoxShape( osgbBullet::asBtVector3( halfLengths ) );
        btTransform trans; trans.setIdentity();
        trans.setOrigin( osgbBullet::asBtVector3( center ) );
        cs->addChildShape( trans, box );
    }
    { // bottom of window -Y
        osg::Vec3 halfLengths( xDim*.5, thick*.5, zDim*.5 );
        osg::Vec3 center( 0., yDim*.5, 0. );
        btBoxShape* box = new btBoxShape( osgbBullet::asBtVector3( halfLengths ) );
        btTransform trans; trans.setIdentity();
        trans.setOrigin( osgbBullet::asBtVector3( center ) );
        cs->addChildShape( trans, box );
    }
    shakeBoxBullet.motion = new osgbBullet::MotionState();
    shakeBoxBullet.motion->setTransform( shakeBoxOSG );
    btScalar mass( 0.0 );
    btVector3 inertia( 0, 0, 0 );
    btRigidBody::btRigidBodyConstructionInfo rb( mass, shakeBoxBullet.motion, shakeBoxBullet.cs, inertia );
    shakeBoxBullet.body = new btRigidBody( rb );
    shakeBoxBullet.body->setCollisionFlags( shakeBoxBullet.body->getCollisionFlags() | btCollisionObject::CF_KINEMATIC_OBJECT );
    shakeBoxBullet.body->setActivationState( DISABLE_DEACTIVATION );
    dynamicsWorld->addRigidBody( shakeBoxBullet.body );
}

void initPhysics()
{
    collisionConfiguration = new btDefaultCollisionConfiguration();
    dispatcher = new btCollisionDispatcher( collisionConfiguration );
    solver = new btSequentialImpulseConstraintSolver;

    btVector3 worldAabbMin( -10000, -10000, -10000 );
    btVector3 worldAabbMax( 10000, 10000, 10000 );
    inter = new btAxisSweep3( worldAabbMin, worldAabbMax, 1000 );

    dynamicsWorld = new btDiscreteDynamicsWorld( dispatcher, inter, solver, collisionConfiguration );

    dynamicsWorld->setGravity( btVector3( 0, 0, 9.8 ) );

    for (int i=0;i<DICE_COUNT;++i)
        initDiePhysic(diceOSG[i], &diceBullet[i]);
    initShakeBoxPhysic();
}

void clearObjectPhysic(struct ObjectPhysic *op)
{
    dynamicsWorld->removeRigidBody(op->body);
    delete op->body;
    delete op->motion;
    delete op->cs;
}

void clearPhysics()
{
    for (int i=0;i<DICE_COUNT;++i)
        clearObjectPhysic(&diceBullet[i]);
    clearObjectPhysic(&shakeBoxBullet);

    delete dynamicsWorld;
    delete inter;
    delete solver;
    delete dispatcher;
    delete collisionConfiguration;
}

class ShakeManipulator : public osgGA::GUIEventHandler
{
public:
    ShakeManipulator( osgbBullet::MotionState* motion )
      : _motion( motion )
    {}

    bool handle( const osgGA::GUIEventAdapter& ea, osgGA::GUIActionAdapter& )
    {
        switch( ea.getEventType() )
        {
            case osgGA::GUIEventAdapter::KEYUP:
            {
                if (ea.getKey()==osgGA::GUIEventAdapter::KEY_Space)
                {
                    btTransform trans; trans.setIdentity();
                    _motion->setWorldTransform( trans );

                    return true;
                }

                return false;
            }

            case osgGA::GUIEventAdapter::PUSH:
            {
                _lastX = ea.getXnormalized();
                _lastY = ea.getYnormalized();

                btTransform world;
                _motion->getWorldTransform( world );
                btVector3 o = world.getOrigin();
                o[ 2 ] = 0.25;
                world.setOrigin( o );
                _motion->setWorldTransform( world );

                return true;
            }
            case osgGA::GUIEventAdapter::DRAG:
            {
                btVector3 move;
                move[ 0 ] = _lastX - ea.getXnormalized();
                move[ 1 ] = ea.getYnormalized() - _lastY;
                move[ 2 ] = 0.;
                move *= 10.;
                btTransform moveTrans; moveTrans.setIdentity();
                moveTrans.setOrigin( move );
                btTransform world;
                _motion->getWorldTransform( world );
                btTransform netTrans = moveTrans * world;
                btVector3 o = netTrans.getOrigin();
                o[ 2 ] = 0.;
                netTrans.setOrigin( o );

                _motion->setWorldTransform( netTrans );

                _lastX = ea.getXnormalized();
                _lastY = ea.getYnormalized();

                return true;
            }
            default:
            break;
        }
        return false;
    }

    osgbBullet::MotionState* _motion;
    float _lastX, _lastY;
};

/* Array for all available keys */
SDL_bool pressedkeys[SDL_NUM_SCANCODES];

/* function to handle key press events */
void handleKeyPress(SDL_keysym* keysym, SDL_bool pressed)
{
    if (pressed==SDL_TRUE)
    {
       pressedkeys[keysym->scancode]=SDL_TRUE;
    }
    else
    {
       pressedkeys[keysym->scancode]=SDL_FALSE;
    }
}

bool convertEvent(SDL_Event& event, osgGA::EventQueue& eventQueue)
{
    switch (event.type) {

        case SDL_MOUSEMOTION:
            eventQueue.mouseMotion(event.motion.x, event.motion.y);
            return true;

        case SDL_MOUSEBUTTONDOWN:
            eventQueue.mouseButtonPress(event.button.x, event.button.y, event.button.button);
            return true;

        case SDL_MOUSEBUTTONUP:
            eventQueue.mouseButtonRelease(event.button.x, event.button.y, event.button.button);
            return true;

        case SDL_KEYUP:
            eventQueue.keyRelease( (osgGA::GUIEventAdapter::KeySymbol) event.key.keysym.unicode);
            return true;

        case SDL_KEYDOWN:
            eventQueue.keyPress( (osgGA::GUIEventAdapter::KeySymbol) event.key.keysym.unicode);
            return true;

        case SDL_VIDEORESIZE:
            eventQueue.windowResize(0, 0, event.resize.w, event.resize.h );
            return true;

        default:
            break;
    }
    return false;
}

void main_loop(int argc, char** argv)
{
    /* These are to calculate our fps */
    static GLint T0=SDL_GetTicks();
    static GLint Frames=0;

    /* main loop variable */
    SDL_bool done=SDL_FALSE;
    /* SDL event */
    SDL_Event event;
    /* whether or not the window is active */
    SDL_bool isActive=SDL_TRUE;

    osg::Group* root = new osg::Group;
    for (int i=0;i<DICE_COUNT;++i)
        root->addChild( diceOSG[i]=makeDie() );

    shakeBoxOSG = makeShakeBox();
    root->addChild( shakeBoxOSG );

    initPhysics();

    osgViewer::Viewer viewer;
    osg::ref_ptr<osgViewer::GraphicsWindowEmbedded> gw = viewer.setUpViewerAsEmbeddedInWindow(0,0,320,240);
    viewer.setSceneData( root );
    viewer.getCamera()->setViewMatrixAsLookAt(
        osg::Vec3( 0, 0, -20 ), osg::Vec3( 0, 0, 0 ), osg::Vec3( 0, 1, 0 ) );
    viewer.getCamera()->setProjectionMatrixAsPerspective( 40., 1., 1., 50. );
    ShakeManipulator *manip = new ShakeManipulator( shakeBoxBullet.motion );
    viewer.addEventHandler( manip );

    viewer.realize();

    double prevSimTime = 0.;

    /* wait for events */
    while (!done)
    {
    	WIZ_Input_Emulation();
        /* handle the events in the queue */
        while (SDL_PollEvent(&event))
        {
            // pass the SDL event into the viewers event queue
            convertEvent(event, *(gw->getEventQueue()));

            switch (event.type)
            {
                case SDL_WINDOWEVENT:
                     switch (event.window.event)
                     {
                         case SDL_WINDOWEVENT_CLOSE:
                              done=SDL_TRUE;
                              break;
                         case SDL_WINDOWEVENT_HIDDEN:
                              isActive=SDL_FALSE;
                              break;
                         case SDL_WINDOWEVENT_SHOWN:
                              isActive=SDL_TRUE;
                              break;
                     }
                     break;
                case SDL_KEYDOWN:
                     /* handle key presses */
                     handleKeyPress(&event.key.keysym, SDL_TRUE);
                     done = pressedkeys[SDL_SCANCODE_ESCAPE];
                     break;
                case SDL_KEYUP:
                     /* handle key presses */
                     handleKeyPress(&event.key.keysym, SDL_FALSE);
                     if (event.key.keysym.scancode == SDL_SCANCODE_RETURN)
                     {
                         clearPhysics();
                         initPhysics();
                         manip->_lastX = 0;
                         manip->_lastY = 0;
                         manip->_motion = shakeBoxBullet.motion;
                         btTransform trans; trans.setIdentity();
                         manip->_motion->setWorldTransform( trans );
                     }
                     break;
                case SDL_QUIT:
                     /* handle quit requests */
                     done=SDL_TRUE;
                     break;
                default:
                     break;
            }
        }
        const double currSimTime = viewer.getFrameStamp()->getSimulationTime();
        double elapsed( currSimTime - prevSimTime );
        if( viewer.getFrameStamp()->getFrameNumber() < 3 )
            elapsed = 1./60.;

        dynamicsWorld->stepSimulation( elapsed, 4, elapsed/4. );
        prevSimTime = currSimTime;
        viewer.frame();

        SDL_GL_SwapWindow(window);

        /* Gather our frames per second */
        Frames++;
        {
            GLint t=SDL_GetTicks();
            if (t-T0>=5000)
            {
                GLfloat seconds=(t-T0)/1000.0;
                GLfloat fps=Frames/seconds;
                printf("%d frames in %g seconds = %g FPS\n", Frames, seconds, fps);
                T0=t;
                Frames=0;
            }
        }
    }
    clearPhysics();
}

