#include "headers.h"

/* function to load in bitmap as a GL texture */
SDL_bool LoadGLTextures(char *name, GLuint* texture, int num_textures)
{
    /* Status indicator */
    SDL_bool status=SDL_FALSE;

    /* Create storage space for the texture */
    SDL_Surface* TextureImage = NULL;

    /* Load The Bitmap, Check For Errors, If Bitmap's Not Found Quit */
    if ((TextureImage=SDL_LoadBMP( name)))
    {
        /* Since OpenGL ES 1.0/1.1 has no BGRA extension we must swap R & B */
        {
            GLuint it;
            GLuint jt;

            for (jt=0; jt<TextureImage->h; jt++)
            {
                for (it=0; it<TextureImage->w; it++)
                {
                    switch (TextureImage->format->BytesPerPixel)
                    {
                        case 3: /* 24 bpp */
                             {
                                 unsigned char rvalue;
                                 unsigned char bvalue;

                                 rvalue=*((unsigned char*)TextureImage->pixels+
                                     jt*TextureImage->pitch+it*TextureImage->format->BytesPerPixel);
                                 bvalue=*((unsigned char*)TextureImage->pixels+
                                     jt*TextureImage->pitch+it*TextureImage->format->BytesPerPixel+2);
                                 /* Convert RGB to BGR and vice versa */
                                 *((unsigned char*)TextureImage->pixels+
                                     jt*TextureImage->pitch+it*TextureImage->format->BytesPerPixel)=bvalue;
                                 *((unsigned char*)TextureImage->pixels+
                                     jt*TextureImage->pitch+it*TextureImage->format->BytesPerPixel+2)=rvalue;
                             }
                             break;
                        case 4: /* 32 bpp */
                             {
                                 unsigned long rgbvalue;

                                 rgbvalue=*(unsigned long*)((unsigned char*)TextureImage->pixels+
                                     jt*TextureImage->pitch+it*TextureImage->format->BytesPerPixel);
                                 /* Convert ARGB to ABGR and vice versa */
                                 rgbvalue=(rgbvalue & 0xFF00FF00) | ((rgbvalue<<16) & 0x00FF0000) |
                                          ((rgbvalue>>16) & 0x000000FF);
                                 *(unsigned long*)((unsigned char*)TextureImage->pixels+
                                     jt*TextureImage->pitch+it*TextureImage->format->BytesPerPixel)=rgbvalue;
                             }
                             break;
                    }
                }
            }
        }

        /* Set the status to true */
        status=SDL_TRUE;

        /* Create The Texture */
        glGenTextures(num_textures, &texture[0]);

        /* Load in texture 1 */
        /* Typical Texture Generation Using Data From The Bitmap */
        glBindTexture(GL_TEXTURE_2D, texture[0]);

        /* Generate The Texture */
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, TextureImage->w,
                     TextureImage->h, 0, GL_RGB, GL_UNSIGNED_BYTE,
                     TextureImage->pixels);

        /* Nearest Filtering */
        glTexParameterx(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameterx(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

        /* Load in texture 2 */
        /* Typical Texture Generation Using Data From The Bitmap */
        glBindTexture(GL_TEXTURE_2D, texture[1]);

        /* Linear Filtering */
        glTexParameterx(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameterx(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

        /* Generate The Texture */
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, TextureImage->w,
                     TextureImage->h, 0, GL_RGB, GL_UNSIGNED_BYTE,
                     TextureImage->pixels);

        /* Load in texture 3 */
        /* Typical Texture Generation Using Data From The Bitmap */
        glBindTexture(GL_TEXTURE_2D, texture[2]);

        /* Mipmapped Filtering */
        glTexParameterx(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
        glTexParameterx(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

        /* Generate The MipMapped Texture */
        gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB, TextureImage->w,
                          TextureImage->h, GL_RGB, GL_UNSIGNED_BYTE,
                          TextureImage->pixels);
    }

    /* Free up any memory we may have used */
    if (TextureImage)
    {
        SDL_FreeSurface(TextureImage);
    }

    return status;
}

