#include "wiz.h"

#ifndef WIZ

int WIZ_Init() { return 0; }
void WIZ_Close() {}
void WIZ_Input_Emulation() {}
unsigned long WIZ_TimerRead() { return 0; }
void WIZ_TimerDelay(unsigned long delay) {}

#else

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <string.h>

#include <SDL13/SDL.h>

#define TRUE	1
#define FALSE	0
#define BIT(number) (1<<(number))

/* GPIO */
#define GPIOAPAD		(memregs16[0xA018>>1])
#define GPIOBPAD		(memregs16[0xA058>>1])
#define GPIOCPAD		(memregs16[0xA098>>1])

enum {
	BTN_A = BIT(20),
	BTN_B = BIT(21),
	BTN_X = BIT(22),
	BTN_Y = BIT(23),
	BTN_L = BIT(7),
	BTN_R = BIT(6),
	BTN_MENU = BIT(9),
	BTN_SELECT = BIT(8),
	BTN_LEFT = BIT(16),
	BTN_RIGHT = BIT(17),
	BTN_UP = BIT(18),
	BTN_DOWN = BIT(19),
	BTN_VOLUP = BIT(10),
	BTN_VOLDOWN = BIT(11)
};

int memfd;
volatile uint32_t *memregs32;
volatile uint16_t *memregs16;

static uint32_t prev_value;

#include "tslib.h"

int mousefd = -2;
struct tsdev *ts_dev;
int _wiz_init_ = 0;

int openTS()
{
	int i;
	const char *mousedev;
	const char *mousedrv;

	mousedev = SDL_getenv("TSLIB_TSDEVICE");
	mousefd = -1;
				
	// jhkang 20080625
	if (1/*mousedrv && (SDL_strcmp(mousedrv, "TSLIB") == 0) */) {
		if (mousedev == NULL) mousedev = strdup("/dev/input/event0");
		if (mousedev != NULL) {
			ts_dev = ts_open(mousedev, 1);
			if ((ts_dev != NULL) && (ts_config(ts_dev) >= 0)) {
				printf("Using tslib touchscreen\n");
				mousefd = ts_fd(ts_dev);
			}
		}
	}
	return mousefd;
}

int button_state = 0, mouse_prev_x = 0, mouse_prev_y = 0;

void fillMotionEvent(SDL_Event *motionevent, struct ts_sample *sample)
{
	motionevent->type = SDL_MOUSEMOTION;
	motionevent->motion.which = 0;
	motionevent->motion.state = SDL_PRESSED;
	motionevent->motion.pressure = 1;
	motionevent->motion.pressure_max = 1;
	motionevent->motion.pressure_min = 0;
	motionevent->motion.cursor = 0;
	motionevent->motion.xrel = sample->x - mouse_prev_x;
	motionevent->motion.yrel = sample->y - mouse_prev_y;
	motionevent->motion.x = sample->x;
	motionevent->motion.y = sample->y;
}

void WIZ_Mouse_Emulation()
{
	struct ts_sample sample;
	int state_changed;
	int i, motion=0;
	Uint8 state;
	SDL_Event fakeevent;
	SDL_Event motionevent;

	while (ts_read(ts_dev, &sample, 1) > 0)
	{
		/* See what changed */
		state = sample.pressure > 0 ? 1 : 0;
		if ( (button_state & 1) != state )
		{
			motion = 0;
			if ( state )
			{
				fakeevent.type = SDL_MOUSEBUTTONDOWN;
				fakeevent.button.state = SDL_PRESSED;
			}
			else
			{
				fillMotionEvent(&motionevent, &sample);
				if (motionevent.motion.xrel || motionevent.motion.yrel)
			        	SDL_PushEvent(&motionevent);
				fakeevent.type = SDL_MOUSEBUTTONUP;
				fakeevent.button.state = SDL_RELEASED;
			}
			fakeevent.button.which = 0;
			fakeevent.button.button = 1;
			mouse_prev_x = fakeevent.button.x = sample.x;
			mouse_prev_y = fakeevent.button.y = sample.y;
	        	SDL_PushEvent(&fakeevent);
			button_state = state;
			break;
		}
		else if (state)
		{
			fillMotionEvent(&motionevent, &sample);
			motion = 1;
		}
	}
	if (motion && (motionevent.motion.xrel || motionevent.motion.yrel))
	{
		mouse_prev_x = motionevent.motion.x;
		mouse_prev_y = motionevent.motion.y;
        	SDL_PushEvent(&motionevent);
	}
	return;
}

#define TIMER_BASE3 0x1980
#define TIMER_REG(x) memregs32[(TIMER_BASE3 + x) >> 2]

void WIZ_TimerInit( void )
{
	TIMER_REG(0x44) = 0x922;
	TIMER_REG(0x40) = 0x0c;
	TIMER_REG(0x08) = 0x6b;
}

void WIZ_TimerClose(void)
{
	TIMER_REG(0x40) = 0x0c;
	TIMER_REG(0x08) = 0x23;
	TIMER_REG(0x00) = 0;
	TIMER_REG(0x40) = 0;
	TIMER_REG(0x44) = 0;
}

unsigned long WIZ_TimerRead(void)
{
	unsigned int microsec;
	if (!_wiz_init_)
		return 0;        
	TIMER_REG(0x08) = 0x4b;  /* run timer, latch value */
	microsec = TIMER_REG(0);
	return microsec;
}

void WIZ_TimerDelay(unsigned long delay)
{
	unsigned int start;
	if (!_wiz_init_)
		return;        
	start = WIZ_TimerRead();
	while(WIZ_TimerRead()-start < delay) {}
}

void WIZ_Close( void )
{
	if (memfd>=0)
	{
		if(memregs32 != (volatile uint32_t*)0xFFFFFFFF)
		{
			WIZ_TimerClose();
			munmap((void *)memregs32, 0x20000);
		}
		close(memfd);
	}
	memregs16 = NULL;
	memregs32 = NULL;
	if (mousefd>=0)
	{
		ts_close(ts_dev);
	}
}

int WIZ_Init( void )
{
	/* open /dev/mem to access registers */
	memfd = open("/dev/mem", O_RDWR);
	if(memfd < 0) {
		printf("Could not open /dev/mem!\n");
		WIZ_Close();
		return 1;
	}

	/* get access to the registers */
	memregs32 = (volatile uint32_t*)mmap(0, 0x20000, PROT_READ|PROT_WRITE, MAP_SHARED, memfd, 0xC0000000);
	if(memregs32 == (volatile uint32_t*)0xFFFFFFFF) {
		printf("Could not mmap hardware registers!\n");
		WIZ_Close();
		return 1;
	}
	memregs16 = (volatile uint16_t*)memregs32;

	WIZ_TimerInit();

	openTS();

	_wiz_init_ = 1;
	return 0;
}

uint32_t getbuttons(void) {
	return ~((GPIOCPAD << 16) | GPIOBPAD);
}

void WIZ_Key_Emulation( char prev_press, char curr_press, SDL_scancode scan_code )
{
	SDL_Event fakeevent;

	if( prev_press == FALSE && curr_press == TRUE ) {
	      fakeevent.type			= SDL_KEYDOWN;
	      fakeevent.key.state		= SDL_PRESSED;
	      fakeevent.key.type		= SDL_KEYDOWN;
	      fakeevent.key.keysym.mod		= KMOD_NONE;
	      fakeevent.key.keysym.scancode	= scan_code;
	      fakeevent.key.keysym.sym		= SDL_SCANCODE_TO_KEYCODE(scan_code);
	      SDL_PushEvent(&fakeevent);
	}
	else if( prev_press == TRUE && curr_press == FALSE ) {
	      fakeevent.type			= SDL_KEYUP;
	      fakeevent.key.state		= SDL_RELEASED;
	      fakeevent.key.type		= SDL_KEYUP;
	      fakeevent.key.keysym.mod		= KMOD_NONE;
	      fakeevent.key.keysym.scancode	= scan_code;
	      fakeevent.key.keysym.sym		= SDL_SCANCODE_TO_KEYCODE(scan_code);
	      SDL_PushEvent(&fakeevent);
	}
}

void WIZ_Keyboard_Emulation( void )
{
	uint32_t curr_value;
	char prev_press;
	char curr_press;

	curr_value = getbuttons();

	// DPAD
	if( curr_value & BTN_L ) {
		curr_press = (curr_value & BTN_UP) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_UP) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_UP );

		curr_press = (curr_value & BTN_DOWN) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_DOWN) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_DOWN );

		curr_press = (curr_value & BTN_LEFT) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_LEFT) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_LEFT );

		curr_press = (curr_value & BTN_RIGHT) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_RIGHT) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_RIGHT );
	} else {
		curr_press = (curr_value & BTN_UP) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_UP) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_KP_8 );

		curr_press = (curr_value & BTN_DOWN) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_DOWN) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_KP_2 );

		curr_press = (curr_value & BTN_LEFT) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_LEFT) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_KP_4 );

		curr_press = (curr_value & BTN_RIGHT) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_RIGHT) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_KP_6 );
	}

	if( curr_value & BTN_L ) {
		curr_press = (curr_value & BTN_A) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_A) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_F );

		curr_press = (curr_value & BTN_B) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_B) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_B );

		curr_press = (curr_value & BTN_X) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_X) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_G );

		curr_press = (curr_value & BTN_Y) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_Y) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_T );

		curr_press = (curr_value & BTN_R) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_R) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_HOME );

		curr_press = (curr_value & BTN_VOLDOWN) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_VOLDOWN) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_PAGEUP );

		curr_press = (curr_value & BTN_VOLUP) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_VOLUP) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_PAGEDOWN );
	} else {
		curr_press = (curr_value & BTN_A) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_A) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_L );

		curr_press = (curr_value & BTN_B) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_B) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_M );

		curr_press = (curr_value & BTN_X) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_X) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_SPACE );

		curr_press = (curr_value & BTN_Y) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_Y) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_P );

		curr_press = (curr_value & BTN_R) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_R) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_TAB );

		curr_press = (curr_value & BTN_VOLDOWN) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_VOLDOWN) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_KP_PLUS );

		curr_press = (curr_value & BTN_VOLUP) ? TRUE : FALSE;
		prev_press = (prev_value & BTN_VOLUP) ? TRUE : FALSE;
		WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_KP_MINUS );
	}

	curr_press = (curr_value & BTN_MENU) ? TRUE : FALSE;
	prev_press = (prev_value & BTN_MENU) ? TRUE : FALSE;
	WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_ESCAPE );

	curr_press = (curr_value & BTN_SELECT) ? TRUE : FALSE;
	prev_press = (prev_value & BTN_SELECT) ? TRUE : FALSE;
	WIZ_Key_Emulation( prev_press, curr_press, SDL_SCANCODE_RETURN );

	prev_value = curr_value;
} 

void WIZ_Input_Emulation()
{
	WIZ_Keyboard_Emulation();
	if (mousefd>=0)
		WIZ_Mouse_Emulation();
}

#endif // WIZ
