#include "wiz.h"

/* screen width, height, and bit depth */
#define WINDOW_WIDTH  320
#define WINDOW_HEIGHT 240
#define WINDOW_BPP     16

/* SDL window */
SDL_WindowID window = -1;
/* SDL GL ES context, binded to SDL window */
SDL_GLContext glcontext=NULL;

void main_loop(int,char**);

void cleanAll()
{
    /* clean up the window */
    if (glcontext!=NULL)
    {
        SDL_GL_DeleteContext(glcontext);
        glcontext = NULL;
    }
//    if (window!=-1)
//        SDL_DestroyWindow(window);

    WIZ_Close();

    SDL_Quit();
}

/* function to release/destroy our resources and restoring the old desktop */
void Quit(int returnCode)
{
    cleanAll();
    /* and exit appropriately */
    exit(returnCode);
}

/* function to reset our viewport after a window resize */
SDL_bool resizeWindow(int width, int height)
{
    /* Height / width ration */
    GLfloat ratio;

    /* Protect against a divide by zero */
    if (height==0)
    {
        height=1;
    }

    ratio=(GLfloat)width/(GLfloat)height;

    /* Setup our viewport. */
    glViewport(0, 0, (GLint)width, (GLint)height);

    /* change to the projection matrix and set our viewing volume. */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    /* Set our perspective */
    gluPerspective(45.0f, ratio, 0.1f, 100.0f);

    /* Make sure we're chaning the model view and not the projection */
    glMatrixMode(GL_MODELVIEW);

    /* Reset The View */
    glLoadIdentity();

    return SDL_TRUE;
}

int main(int argc, char** argv)
{
    int status;

    if( WIZ_Init() ) Quit(-1);

    status=SDL_Init(SDL_INIT_VIDEO);
    if (status<0)
    {
        fprintf(stderr, "Can't init default SDL video driver: %s\n", SDL_GetError());
        exit(-1);
    }

    // Select first display
    status=SDL_SelectVideoDisplay(0);
    if (status<0)
    {
        fprintf(stderr, "Can't attach to first display: %s\n", SDL_GetError());
        exit(-1);
    }

    // Sets up OpenGL ES double buffering and visual parameters
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
    SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
    SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 6);
    SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);

    window=SDL_CreateWindow("Test OpenSceneGraph",
        SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED,
        WINDOW_WIDTH, WINDOW_HEIGHT,
        SDL_WINDOW_RESIZABLE | SDL_WINDOW_OPENGL | SDL_WINDOW_SHOWN);
    if (window==0)
    {
        fprintf(stderr, "Can't create window: %s\n", SDL_GetError());
        Quit(1);
    }

    glcontext=SDL_GL_CreateContext(window);
    if (glcontext==NULL)
    {
        fprintf(stderr, "Can't create OpenGL ES context: %s\n", SDL_GetError());
        Quit(1);
    }

    status=SDL_GL_MakeCurrent(window, glcontext);
    if (status<0)
    {
        fprintf(stderr, "Can't set current OpenGL ES context: %s\n", SDL_GetError());
        Quit(1);
    }

    // Enable swap on VSYNC
    SDL_GL_SetSwapInterval(1);

    // resize the initial window
    resizeWindow(WINDOW_WIDTH, WINDOW_HEIGHT);

    main_loop(argc,argv);

    // clean ourselves up
    cleanAll();

    return 0;
}

